-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local LibTSMWoW = select(2, ...).LibTSMWoW
local Inbox = LibTSMWoW:Init("API.Inbox")
local ClientInfo = LibTSMWoW:Include("Util.ClientInfo")
local EnumType = LibTSMWoW:From("LibTSMUtil"):Include("BaseType.EnumType")
local MAIL_TYPE = EnumType.New("MAIL_TYPE", {
	SALE = EnumType.NewValue(),
	BUY = EnumType.NewValue(),
	CANCEL = EnumType.NewValue(),
	EXPIRE = EnumType.NewValue(),
	OTHER = EnumType.NewValue(),
	LETTER = EnumType.NewValue(),
})
Inbox.MAIL_TYPE = MAIL_TYPE



-- ============================================================================
-- Module Functions
-- ============================================================================

---Gets the number of inbox mail items.
---@return number numItems
---@return number totalItems
function Inbox.GetNumItems()
	return GetInboxNumItems()
end

---Gets header info.
---@param index number The mail index
---@return number texture
---@return string sender
---@return string subject
---@return number money
---@return number cod
---@return number daysLeft
---@return number numItems
---@return boolean textCreated
---@return boolean isGM
---@return boolean canReply
function Inbox.GetHeaderInfo(index)
	local _, texture, sender, subject, money, cod, daysLeft, numItems, _, _, textCreated, canReply, isGM = GetInboxHeaderInfo(index)
	return texture, sender, subject, money, cod, daysLeft, numItems, textCreated, isGM, canReply
end

---Gets invoice info.
---@param index number The mail index
---@return string? invoiceType
---@return string? itemName
---@return string? playerName
---@return number bid
---@return number buyout
---@return number deposit
---@return number consignment
---@return number? etaHour
---@return number? etaMin
---@return number? count
function Inbox.GetInvoiceInfo(index)
	local invoiceType, itemName, playerName, bid, buyout, deposit, consignment, _, etaHour, etaMin, count = GetInboxInvoiceInfo(index)
	return invoiceType, itemName, playerName, bid, buyout, deposit, consignment, etaHour, etaMin, count
end

---Gets the crafting order info for a mail.
---@param index number The mail index
---@return Enum.RcoCloseReason? reason
---@return number? commissionPaid
---@return string? customerName
---@return string? crafterName
---@return string? recipeName
function Inbox.GetCraftingOrderInfo(index)
	if not select(4, Inbox.GetText(index)) then
		return nil, nil, nil, nil, nil
	end
	local info = C_Mail.GetCraftingOrderMailInfo(index)
	return info.reason, info.commissionPaid, info.customerName, info.crafterName, info.recipeName
end

---Gets the type of a mail.
---@param index index The mail index
---@return EnumValue
function Inbox.GetMailType(index)
	local _, _, subject, money, cod, _, numItems, _, isGM = Inbox.GetHeaderInfo(index)
	if isGM or (cod and cod > 0) or (money == 0 and (not numItems or numItems == 0)) then
		return MAIL_TYPE.LETTER
	end

	local invoiceType = Inbox.GetInvoiceInfo(index)
	if money and money > 0 and invoiceType == "seller" then
		return MAIL_TYPE.SALE
	elseif numItems and numItems > 0 and invoiceType == "buyer" then
		return MAIL_TYPE.BUY
	elseif invoiceType then
		return MAIL_TYPE.OTHER
	elseif strfind(subject, gsub("^"..AUCTION_REMOVED_MAIL_SUBJECT, "%%s", "")) then
		return MAIL_TYPE.CANCEL
	elseif strfind(subject, gsub("^"..AUCTION_EXPIRED_MAIL_SUBJECT, "%%s", "")) then
		return MAIL_TYPE.EXPIRE
	elseif ClientInfo.HasFeature(ClientInfo.FEATURES.CRAFTING_ORDERS) then
		local craftingOrderReason = Inbox.GetCraftingOrderInfo(index)
		if craftingOrderReason == Enum.RcoCloseReason.RcoCloseCancel then
			return MAIL_TYPE.CANCEL
		elseif craftingOrderReason == Enum.RcoCloseReason.RcoCloseExpire then
			return MAIL_TYPE.EXPIRE
		elseif craftingOrderReason == Enum.RcoCloseReason.RcoCloseCrafterFulfill then
			return MAIL_TYPE.SALE
		elseif craftingOrderReason == Enum.RcoCloseReason.RcoCloseFulfill then
			return MAIL_TYPE.BUY
		end
	end
	return MAIL_TYPE.OTHER
end

---Gets the text of a mail.
---@param index index The mail index
---@return string body
---@return boolean isTakeable
---@return boolean isInvoice
---@return boolean isConsortium
function Inbox.GetText(index)
	local body, _, _, isTakeable, isInvoice, isConsortium = GetInboxText(index)
	return body, isTakeable, isInvoice, isConsortium
end

---Gets the max number of attachment that can be in a received mail.
---@return number
function Inbox.GetMaxAttachments()
	return ATTACHMENTS_MAX
end

---Gets info on a mail attachment.
---@param index number The mail index
---@param attachIndex number The attachment index
---@return string itemLink
---@return number quantity
function Inbox.GetAttachment(index, attachIndex)
	local _, _, _, stackSize = GetInboxItem(index, attachIndex)
	local itemLink = GetInboxItemLink(index, attachIndex)
	return itemLink, stackSize
end

---Gets the stack size of an item attachment.
---@param index number The mail index
---@param attachIndex number The attachment index
---@return number?
function Inbox.GetAttachmentStackSize(index, attachIndex)
	local _, _, _, stackSize = GetInboxItem(index, attachIndex)
	return stackSize
end

---Gets a mail item link (or a pet itemString if it's a pet).
---@param index number The mail index
---@param attachIndex number The attachment index
---@return string?
function Inbox.GetAttachmentItemLink(index, attachIndex)
	return GetInboxItemLink(index, attachIndex)
end

---Gets the max number of attachment that can be sent.
---@return number
function Inbox.GetMaxSendAttachments()
	return ATTACHMENTS_MAX_SEND
end

---Gets info on a send attachment.
---@param index number The attachment index
---@return string itemLink
---@return number quantity
function Inbox.GetSendAttachment(index)
	local itemLink = GetSendMailItemLink(index)
	local _, _, _, quantity = GetSendMailItem(index)
	return itemLink, quantity
end

---Register a secure hook function for when mail is sent.
---@param func fun(target: string) The function to call
function Inbox.SecureHookSendMail(func)
	hooksecurefunc("SendMail", func)
end

---Register a secure hook function for when mail is returned.
---@param func fun(index: number) The function to call
function Inbox.SecureHookReturnMail(func)
	hooksecurefunc("ReturnInboxItem", func)
end
